/*******************************************************************************/
/*                                                                             */
/*  ip.c, ip calc routines                                                     */
/*  Copyright (C) 2002 Dennis Pries                                            */
/*                                                                             */
/*  This program is free software; you can redistribute it and/or modify       */
/*  it under the terms of the GNU General Public License as published by       */
/*  the Free Software Foundation; either version 2 of the License, or          */
/*  (at your option) any later version.                                        */
/*                                                                             */
/*  This program is distributed in the hope that it will be useful,            */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              */
/*  GNU General Public License for more details.                               */
/*                                                                             */
/*  You should have received a copy of the GNU General Public License          */
/*  along with this program; if not, write to the Free Software                */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */
/*                                                                             */
/*******************************************************************************/

#include	<stdrom.h>
#include	"define.h"
#include	"ip.h"

/* Convert a number to ip.netmask */
/* Return FALSE if number invalid */
bool numToNetmask(byte num, IP *ip)
{
  int i;

  ip->netmask[0] = 0;
  ip->netmask[1] = 0;
  ip->netmask[2] = 0;
  ip->netmask[3] = 0;

  if (num < 1 || num > 32) return(FALSE);

  num = 32 - num;

  for(i=0; i<num; i++)
  {
    if (i < 8)
      ip->netmask[3] |= ((int) pow(2, i));

    if (i >= 8 && i < 16)
      ip->netmask[2] |= ((int) pow(2, i - 8));

    if (i >= 16 && i < 24)
      ip->netmask[1] |= ((int) pow(2, i - 16));

    if (i >= 24)
      ip->netmask[0] |= ((int) pow(2, i - 24));
  }

  ip->netmask[0] ^= 0xff;
  ip->netmask[1] ^= 0xff;
  ip->netmask[2] ^= 0xff;
  ip->netmask[3] ^= 0xff;
  return(TRUE);
}

/* Convert ip.netmask to a number */
/* Return FALSE if netmask invalid */
/* num can be NULL to check netmask only */
bool netmaskToNum(IP ip, byte *num)
{
  int i;
  bool zeroseen = FALSE;

  if (num != NULL) *num = 0;
  ip.netmask[0] ^= 0xff;
  ip.netmask[1] ^= 0xff;
  ip.netmask[2] ^= 0xff;
  ip.netmask[3] ^= 0xff;

  for(i=0; i<32; i++)
  {
    if (i < 8)
      if ((ip.netmask[3] & ((int) pow(2, i))) == 0)
      {
        zeroseen = TRUE;
        if (num != NULL) *num = *num + 1;
      } else {
        if (zeroseen == TRUE) return(FALSE);
      }

    if (i >= 8 && i < 16)
      if ((ip.netmask[2] & ((int) pow(2, i - 8))) == 0)
      {
        zeroseen = TRUE;
        if (num != NULL) *num = *num + 1;
      } else {
        if (zeroseen == TRUE) return(FALSE);
      }

    if (i >= 16 && i < 24)
      if ((ip.netmask[1] & ((int) pow(2, i - 16))) == 0)
      {
        zeroseen = TRUE;
        if (num != NULL) *num = *num + 1;
      } else {
        if (zeroseen == TRUE) return(FALSE);
      }

    if (i >= 24)
      if ((ip.netmask[0] & ((int) pow(2, i - 24))) == 0)
      {
        zeroseen = TRUE;
        if (num != NULL) *num = *num + 1;
      } else {
        if (zeroseen == TRUE) return(FALSE);
      }
  }

  return(TRUE);
}

/* Get the network address from an ip/netmask combination */
void getNetwork(IP ip, IP *network)
{
  int i;

  for(i=0; i<4; i++)
  {
    network->ip[i] = ip.ip[i] & ip.netmask[i];
    network->netmask[i] = ip.netmask[i];
  }
}

/* Get network with number n (first = 0) */
/* total is used for speed optimizing in loops */
float getNetworkN(IP base, IP ip, IP *network, float n, float total)
{
  IP tmp = {};
  int i;

  for(i=0; i<4; i++)
  {
    tmp.ip[i] = base.ip[i];
    tmp.netmask[i] = ip.netmask[i];
  }

  if (total == 0.0)
    total = totalAddresses(tmp);
    
  ipAdd(&tmp, total*n);

  for(i=0; i<4; i++)
  {
    network->ip[i] = tmp.ip[i];
    network->netmask[i] = tmp.netmask[i];
  }
  
  return(total);
}

/* Get the broadcast address from an ip/netmask combination */
void getBroadcast(IP ip, IP *broadcast)
{
  int i;

  for(i=0; i<4; i++)
  {
    broadcast->ip[i] = (ip.ip[i] & ip.netmask[i]) | ip.netmask[i] ^ 0xff;
    broadcast->netmask[i] = ip.netmask[i];
  }
}

/* Get number of total addresses from an ip/netmask combination */
float totalAddresses(IP ip)
{
  return(
    shiftleft((ip.netmask[0] ^ 0xff), 24) +
    shiftleft((ip.netmask[1] ^ 0xff), 16) +
    shiftleft((ip.netmask[2] ^ 0xff), 8)  +
    (ip.netmask[3] ^ 0xff)                +
    1
  );
}

/* Get number of total networks from 2 netmasks */
float totalNetworks(IP base, IP subnet)
{
  int tmp;
  int i;
  int bits = 0;  

  for(i=0; i<4; i++)
  {
	tmp = base.netmask[i] ^ subnet.netmask[i];
	
	if ((tmp & 0x01) != 0) bits++;
	if ((tmp & 0x02) != 0) bits++;
	if ((tmp & 0x04) != 0) bits++;
	if ((tmp & 0x08) != 0) bits++;
	if ((tmp & 0x10) != 0) bits++;
	if ((tmp & 0x20) != 0) bits++;
	if ((tmp & 0x40) != 0) bits++;
	if ((tmp & 0x80) != 0) bits++;
  }
	
  return pow(2, bits);
}

/* Own bit left shift routine with float output */
float shiftleft(byte b, byte num)
{
  byte bittable[8];
  float tmp;
  int i;

  bittable[0] = (b & 0x01);
  bittable[1] = (b & 0x02);
  bittable[2] = (b & 0x04);
  bittable[3] = (b & 0x08);
  bittable[4] = (b & 0x10);
  bittable[5] = (b & 0x20);
  bittable[6] = (b & 0x40);
  bittable[7] = (b & 0x80);

  tmp = 0;
  for(i=0; i<8; i++)
    if(bittable[i] != 0)
     tmp = tmp + pow(2, num + i);

  return(tmp);
}


/* go num ip's higher */
void ipAdd(IP *ip, float num)
{
  int a[4];
  int i;

  a[0] = (int)    (num / 16777216.0);
  num  =      fmod(num, 16777216.0);
  a[1] = (int)    (num / 65536.0);
  num  =      fmod(num, 65536.0);
  a[2] = (int)    (num / 256.0);
  a[3] = (int) fmod(num, 256.0);

  for(i=3; i>0; i--)
  {
    num = (float) ip->ip[i] + (float) a[i];
    if (num < 256.0)
    {
      ip->ip[i] = (int) num;
    } else {
      ip->ip[i] = (int) fmod(num, 256.0);
      ip->ip[i-1]++;
    }
  }

  num = (float) ip->ip[0] + (float) a[0];
  if (num < 256.0)
  {
    ip->ip[0] = (int) num;
  } else {
    /* Overflow */
    for(i=0; i<4; i++)
      ip->ip[i] = 255;
  }
}

/* Convert an ip/netmask combination to a string */
/* Format can be:                                */
/* IP_IPMASK for nnn.nnn.nnn.nnn/mmm.mmm.mmm.mmm */
/* IP_IP     for nnn.nnn.nnn.nnn                 */
/* IP_MASK   for mmm.mmm.mmm.mmm                 */
bool ipToString(IP ip, char *ipnum, int format)
{
  switch(format)
  {
    case IP_IP:
	{
      sprintf(ipnum, "%d.%d.%d.%d", ip.ip[0], ip.ip[1], ip.ip[2], ip.ip[3]);
    }
	break;

    case IP_MASK:
    {
      sprintf(ipnum, "%d.%d.%d.%d", ip.netmask[0], ip.netmask[1], ip.netmask[2], ip.netmask[3]);
    }
    break;

    default:
    {
      sprintf(ipnum, "%d.%d.%d.%d/%d.%d.%d.%d", ip.ip[0], ip.ip[1], ip.ip[2], ip.ip[3],
	                                            ip.netmask[0], ip.netmask[1],
	                                            ip.netmask[2], ip.netmask[3]);
	}
	break;
  }
}


/* Convert a string into an ip-structure */
/* For results see errorToString()       */
int stringToIp(char *ipnum, IP *ip)
{
  int i=0, j=0, k=0, l=0;
  char Buffer[4];
  char *tmpip;
  char *tmpmask;
  char tmpc[33];
  int tmp;

  strcpy(tmpc, ipnum);

  if (tmpc[0] == '\0') return(ERR_FIELDEMTPY);
  if (strstr(tmpc, "/") == NULL) return(ERR_NETMASKMISSED);

  tmpip = strtok(tmpc, "/");
  tmpmask = strtok(NULL, "/");

  while (tmpip[i] != '\0')
  {
    if (tmpip[i] != '.')
    {
      if (j > 3) return(ERR_IPINVALID);
      Buffer[j] = tmpip[i];
      j++;
    } else {
      if (j > 0)
      {
        Buffer[j] = '\0';
        tmp = atoi(Buffer);
        if ((tmp > 255) || (l > 3)) return(ERR_IPINVALID);
        ip->ip[l] = tmp;
      } else {
        ip->ip[l] = 0;
      }
      k++;
      l++;
      j=0;
      Buffer[j] = '\0';
    }
    i++;
  }
  Buffer[j] = '\0';
  tmp = atoi(Buffer);
  if ((ip->ip[l] > 255) || (l < 3)) return(ERR_IPINVALID);
  ip->ip[l] = tmp;

  if (strstr(tmpmask, ".") == NULL)
    if (numToNetmask(atoi(tmpmask), ip))
    {
     return(ERR_NOERROR);
    } else {
     return(ERR_NETMASKINVALID);
    }

  i=0; j=0; k=0; l=0;
  while (tmpmask[i] != '\0')
  {
    if (tmpmask[i] != '.')
    {
      if (j > 3) return(ERR_NETMASKINVALID);
      Buffer[j] = tmpmask[i];
      j++;
    } else {
      if (l > 3) return(ERR_NETMASKINVALID);
      if (j > 0)
      {
        Buffer[j] = '\0';
        tmp = atoi(Buffer);
        if (tmp > 255) return(ERR_NETMASKINVALID);
        ip->netmask[l] = tmp;
      } else {
        ip->ip[l] = 0;
      }
      k++;
      l++;
      j=0;
      Buffer[j] = '\0';
    }
    i++;
  }

  if (l < 3) return(ERR_NETMASKINVALID);

  if (j > 0)
  {
    Buffer[j] = '\0';
    tmp = atoi(Buffer);
    if (ip->netmask[l] > 255) return(ERR_NETMASKINVALID);
    ip->netmask[l] = tmp;
  } else {
    ip->netmask[l] = 0;
  }

  if(netmaskToNum(*ip, NULL))
  {
    return(ERR_NOERROR);
  } else {
    return(ERR_NETMASKINVALID);
  }
}

/* Converts an errorcode to a string */
char* errorToStr(int errcode)
{
  switch(errcode)
  {
    case ERR_NOERROR:
      return("No error");
    break;

    case ERR_FIELDEMTPY:
      return("Field empty");
    break;

    case ERR_IPINVALID:
      return("IP invalid");
    break;

    case ERR_NETMASKMISSED:
      return("Netmask missed");
    break;

    case ERR_NETMASKINVALID:
      return("Netmask invalid");
    break;

	default:
	  return("Unknown error");
	break;
  };
}